import {
    prop,
    getModelForClass,
    modelOptions,
    index,
    plugin,
    mongoose
} from '@typegoose/typegoose';
import {dbconn} from '../decorators/dbconn';
// @ts-ignore
import findOrCreate from 'mongoose-findorcreate';
import {
    Base,
    FindOrCreate,
    TimeStamps
} from "@typegoose/typegoose/lib/defaultClasses";
import {Card} from "./subdoc/Card";
import {Hero} from "./subdoc/Hero";
import {ZError} from "../common/ZError";
import {MoneyTypeConst} from "../constants/MoneyTypeConst";

interface AccountClass extends Base<string>, TimeStamps {}
@dbconn()
// @index({ _id: 1}, { unique: true })
@plugin(findOrCreate)
@modelOptions({schemaOptions:
        {collection: "account", timestamps: true}
})
class UserClass extends FindOrCreate{
    @prop()
    public _id: string;

    @prop({default: false})
    public locked: boolean;
    @prop()
    public lockedtime?: number;

    @prop()
    public comment?: string;
    @prop()
    public lastLogin?: Date;
    /**
     * 已获得卡牌信息
     */
    @prop({ type: () => [Number] })
    public cards: number[];
    /**
     * 已解锁英雄信息
     */
    @prop({ type: Hero, default: new Map() })
    public heros: Map<string, Hero>;
    /**
     * 货币信息
     * coin: 金币
     * diamond: 钻石
     * hero_shard: 通用英雄碎片
     * hero_exp: 通用英雄经验
     * hero_shard_heroid: 英雄专用碎片
     * hero_exp_heroid: 英雄专用经验
     * card_scroll: 抽卡卷轴
     */
    @prop({ type: Number, default: new Map() })
    public moneys: Map<string, number>;


    /**
     * 匹配: 胜利场数, 失败场数, 平局场数, 掉线场数
     */
    @prop({type: Number, default: [0,0,0,0]})
    public normal_stat: number[];
    /**
     * 当前赛季: 胜利场数, 失败场数, 平局场数, 掉线场数
     */
    @prop({type: Number, default: [0,0,0,0]})
    public season_stat: number[];

    /**
     * 所有未定义的信息, 供扩展
     */
    @prop({type: mongoose.Schema.Types.Mixed})
    public extinfo: any;

    public toJson() {
        return {
            accountid: this._id,
        }
    }

    public async unlockHero(heroid: number) {
        if (this.heros.has(heroid + '')) {
            if (!this.heros.get(heroid + '').trial) {
                throw new ZError(102, '你已经解锁了该英雄');
            }
        }
        let money0 = MoneyTypeConst.getHeroShard(heroid);
        let money1 = MoneyTypeConst.HERO_SHARD;
        let count0 = this.moneys.has(money0) ? this.moneys.get(money0) : 0;
        let count1 = this.moneys.has(money1) ? this.moneys.get(money1) : 0;
        // TODO:: 根据配置获取解锁英雄需要的碎片数量
        let needCount = 30;
        if (count0 + count1 < needCount) {
            throw new ZError(102, '碎片数量不足');
        }
        if (count0 > 0) {
            let rest = needCount - count0;
            if (rest >= 0) {
                needCount = rest;
                this.moneys.set(money0, 0);
            } else {
                needCount = 0;
                this.moneys.set(money0, Math.abs(rest));
            }
        }
        // 上面已经过滤了 count0 + count1 > needcount, 所以这里不需要判断碎片是否足够
        if (needCount > 0) {
            count1 = count1 - needCount;
            this.moneys.set(money1, count1);
        }
        let hero = new Hero();
        hero.heroid = heroid;
        hero.free = false;
        hero.trial = false;
        hero.level = 1;
        hero.trial_expire = 0;
        hero.exp = 0;
        this.heros.set(heroid + '', hero);
        return hero;
    }

    public async tryHero(heroid: number) {
        if (this.heros.has(heroid + '')) {
            throw new ZError(102, '你已经解锁了该英雄');
        }
        //TODO:: 根据配置查看该英雄是否可使用
        let hero = new Hero();
        hero.heroid = heroid;
        hero.free = false;
        hero.trial = true;
        hero.level = 1;
        hero.trial_expire = 0;
        hero.exp = 0;
        this.heros.set(heroid+'', hero);
        return hero;
    }

    /**
     * 升级英雄
     * @param heroid
     * @param items
     */
    public async updateHero(heroid: number, items: string) {
        if (items.indexOf(':') < 0) {
            throw new ZError(105, 'items参数格式错误');
        }
        if (!this.heros.has(heroid + '')) {
            throw new ZError(109, '升级英雄不存在');
        }
        let hero = this.heros.get(heroid + '');
        if (hero.trial) {
            throw new ZError(110, '试玩英雄不可升级');
        }

        let moneys: [string, number][] = [];
        if (items.indexOf('|')) {
            let arr = items.split('|');
            for (let str of arr) {
                if(!str) {
                    continue;
                }
                if (str.indexOf(':') < 0) {
                    throw new ZError(105, 'items参数格式错误');
                }

                let subarr = str.split(':');
                if (!MoneyTypeConst.checkMoney(subarr[0])) {
                    throw new ZError(106, `货币不存在: ${subarr[0]}`);
                }
                moneys.push([subarr[0], parseInt(subarr[1])]);
            }
        } else {
            let arr = items.split(':');
            if (!MoneyTypeConst.checkMoney(arr[0])) {
                throw new ZError(106, `货币不存在: ${arr[0]}`);
            }
            moneys.push([arr[0], parseInt(arr[1])]);
        }
        const expPreLvl = 100;
        let totalExp = 0;
        for (let data of moneys) {
            let money = data[0];
            if (MoneyTypeConst.isHeroMoney(money)) {
                let tmpHeroId = MoneyTypeConst.getHeroId(money);
                if (tmpHeroId != heroid) {
                    throw new ZError(107, `指定英雄的货币与当前英雄id不匹配: ${money}`);
                }
            }
            let count = this.moneys.get(money) || 0;
            if (count < data[1]) {
                throw new ZError(108, `当前货币不够: ${money}, in: ${data[1]}, had: ${count}`);
            }
            this.moneys.set(money, count - data[1]);
            // TODO:: 根据配表计算代币->经验值的换算比例
            let exp = data[1];
            totalExp += exp;
        }
        // TODO:: 根据配表读取英雄升级需要的经验值
        let lvlInc = (hero.exp + totalExp) / expPreLvl | 0;
        let expRemain = (hero.exp + totalExp)  % expPreLvl;
        hero.level += lvlInc;
        hero.exp = expRemain;
        return {
            exp: hero.exp,
            level: hero.level,
            exp_gain: totalExp
        }
    }
}


export const User = getModelForClass(UserClass,
// @ts-ignore
    {existingConnection: UserClass['db']});
